#ifndef SCHEDULEDUPDATEVIEW_H
#define SCHEDULEDUPDATEVIEW_H

#include <QObject>
#include <QPointer>
#include <QMap>
#include <QQueue>
#include <QModelIndex>

#include <functional>

class QTimer;
class QShowEvent;
class QResizeEvent;
class QAbstractItemView;

namespace FlyFM
{

class ScheduledViewUpdater;

class ScheduledUpdateViewCommon
{
    friend class ScheduledViewUpdater;
protected:
    virtual ~ScheduledUpdateViewCommon();

    virtual void setUpdater(QObject *context, std::function<void()> updater);
    virtual QModelIndex callIndexBelow(const QModelIndex &index) const;
    virtual void callDataChanged(const QModelIndex &topLeft, const QModelIndex &bottomRight,
                                 const QVector<int> &roles = QVector<int>()) = 0;
};

template <class T>
class ScheduledUpdateView : public T, public ScheduledUpdateViewCommon
{
public:
    ScheduledUpdateView(QWidget *parent = nullptr);

    void scheduleUpdating();

    void setModel(QAbstractItemModel *model) override;
protected:
    void scrollContentsBy(int dx, int dy) override;
    void showEvent(QShowEvent *) override;
    void resizeEvent(QResizeEvent *event) override;

    void rowsAboutToBeRemoved(const QModelIndex &parent, int start, int end) override;
    void rowsInserted(const QModelIndex &parent, int start, int end) override;
private:
    ScheduledViewUpdater *m_viewUpdater;
};

class ScheduledViewUpdater : public QObject
{
    Q_OBJECT
public:
    ScheduledViewUpdater(QAbstractItemView *view);
    ~ScheduledViewUpdater();

    void scheduleUpdating();
    void setModel(QAbstractItemModel *model);

private slots:
    void updateVisibleIndexes();
    void updatePendingIndexes();

    void layoutChanged();
private:
    QTimer *m_updateTimer;
    QQueue<QPersistentModelIndex> pendingIndexes;
    QPointer<QAbstractItemView> m_view;
    ScheduledUpdateViewCommon *m_viewCommon;
};

template <class T>
ScheduledUpdateView<T>::ScheduledUpdateView(QWidget *parent)
    : T(parent)
{
    m_viewUpdater = new ScheduledViewUpdater(this);
}

template <class T>
void ScheduledUpdateView<T>::scheduleUpdating()
{
    m_viewUpdater->scheduleUpdating();
}

template <class T>
void ScheduledUpdateView<T>::setModel(QAbstractItemModel *newModel)
{
    m_viewUpdater->setModel(newModel);

    T::setModel(newModel);
}

template <class T>
void ScheduledUpdateView<T>::scrollContentsBy(int dx, int dy)
{
    T::scrollContentsBy(dx, dy);
    scheduleUpdating();
}

template <class T>
void ScheduledUpdateView<T>::showEvent(QShowEvent *e)
{
    T::showEvent(e);
    scheduleUpdating();
}

template <class T>
void ScheduledUpdateView<T>::resizeEvent(QResizeEvent *event)
{
    T::resizeEvent(event);
    scheduleUpdating();
}

template <class T>
void ScheduledUpdateView<T>::rowsAboutToBeRemoved(const QModelIndex &parent, int start, int end)
{
    T::rowsAboutToBeRemoved(parent, start, end);
    scheduleUpdating();
}

template <class T>
void ScheduledUpdateView<T>::rowsInserted(const QModelIndex &parent, int start, int end)
{
    T::rowsInserted(parent, start, end);
    scheduleUpdating();
}

}

#endif // SCHEDULEDUPDATEVIEW_H
